import itertools
import logging
from collections import defaultdict
from collections.abc import Iterable

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

# TODO: this is a hotfix for changes in pandas 0.25.0, per discussions on the
# pandas-dev listserv, we should try to ask if matplotlib would make it a
# standard feature in their library
from pyam._style import _get_standard_colors
from pyam.exceptions import raise_data_error
from pyam.figures import sankey
from pyam.index import get_index_levels
from pyam.run_control import run_control
from pyam.str import is_str
from pyam.timeseries import cross_threshold
from pyam.utils import (
    IAMC_IDX,
    META_IDX,
    SORT_IDX,
    YEAR_IDX,
    to_list,
)

logger = logging.getLogger(__name__)

# line colors, markers, and styles that are cycled through when not
# explicitly declared
_DEFAULT_PROPS = None

# maximum number of labels after which do not show legends by default
MAX_LEGEND_LABELS = 13

# default legend kwargs for putting legends outside of plots
OUTSIDE_LEGEND = {
    "right": dict(loc="center left", bbox_to_anchor=(1.0, 0.5)),
    "bottom": dict(loc="upper center", bbox_to_anchor=(0.5, -0.2), ncol=3),
}

PYAM_COLORS = {
    # AR6 colours originally from https://github.com/IPCC-WG1/colormaps
    # Final values are used as communicated from the TSU and documented
    # in: https://github.com/IAMconsortium/pyam/pull/566
    # where each file is processed to generate hex values, e.g.:
    # with open('rcp_cat.txt') as f:
    #   for l in f.readlines():
    #     rgb = np.array([int(x) for x in l.strip().split()]) / 256
    #     print(matplotlib.colors.rgb2hex(rgb))
    "AR6-SSP1": "#1e9583",
    "AR6-SSP2": "#4576be",
    "AR6-SSP3": "#f11111",
    "AR6-SSP4": "#e78731",
    "AR6-SSP5": "#8036a7",
    "AR6-SSP1-1.9": "#00a9cf",
    "AR6-SSP1-2.6": "#003466",
    "AR6-SSP2-4.5": "#f69320",
    "AR6-SSP3-7.0": "#df0000",
    "AR6-SSP3-LowNTCF": "#e61d25",
    "AR6-SSP4-3.4": "#2274ae",
    "AR6-SSP4-6.0": "#b0724e",
    "AR6-SSP5-3.4-OS": "#92397a",
    "AR6-SSP5-8.5": "#980002",
    "AR6-RCP-2.6": "#003466",
    "AR6-RCP-4.5": "#709fcc",
    "AR6-RCP-6.0": "#c37900",
    "AR6-RCP-8.5": "#980002",
    # AR5 colours from
    # https://tdaviesbarnard.co.uk/1202/ipcc-official-colors-rcp/
    "AR5-RCP-2.6": "#0000FF",
    "AR5-RCP-4.5": "#79BCFF",
    "AR5-RCP-6.0": "#FF822D",
    "AR5-RCP-8.5": "#FF0000",
    # AR6 WG III scenario categorization (C)
    "AR6-C1": "#97CEE4",  # C1: scenarios that limit warming to 1.5°C (>50%) with no or limited overshoot  # noqa: E501
    "AR6-C2": "#778663",  # C2: scenarios that return warming to 1.5°C (>50%) after a high overshoot # noqa: E501
    "AR6-C3": "#6F7899",  # C3: scenarios that limit warming to 2°C (>67%)
    "AR6-C4": "#A7C682",  # C4: scenarios that limit warming to 2°C (>50%)
    "AR6-C5": "#8CA7D0",  # C5: scenarios that limit warming to 2.5°C (>50%)
    "AR6-C6": "#FAC182",  # C6: scenarios that limit warming to 3°C (>50%)
    "AR6-C7": "#F18872",  # C7: scenarios that limit warming to 4°C (>50%)
    "AR6-C8": "#BD7161",  # C8: scenarios that exceed 4°C warming (≥50%)
    # AR6 Illustrative Mitigation Pathway (IMP)
    "AR6-IMP-LD": "#4FA7BF",  # LD: emphasis on a low demand for resources
    "AR6-IMP-Ren": "#2B7C8B",  # Ren: emphasis on renewables resources
    "AR6-IMP-SP": "#004D52",  # SP: emphasis on sustainable development
    "AR6-IMP-Neg": "#84A12B",  # Neg: emphasis on deployment of carbon dioxide removal
    "AR6-IMP-GS": "#6E7895",  # GS: less rapid near-term mitigation followed by a gradual strengthening # noqa: E501
    # AR6 Illustrative Pathway (IP)
    "AR6-IP-ModAct": "#F29424",  # pathway with moderate climate action
    "AR6-IP-CurPol": "#E31F2B",  # pathway modelling current policies
}


class PlotAccessor:
    """Make plots of IamDataFrame instances"""

    def __init__(self, df):
        self._parent = df

        # assign plotting functions as attributes
        PLOT_MAPPING = {
            "line": line,
            "bar": bar,
            "stack": stack,
            "box": box,
            "pie": pie,
            "scatter": scatter,
            "sankey": sankey,
        }
        # inherit the docstring from the plot function
        for name, func in PLOT_MAPPING.items():
            getattr(self, name).__func__.__doc__ = func.__doc__

    def __call__(self, kind="line", *args, **kwargs):
        return getattr(self, kind)(**kwargs)

    def line(self, **kwargs):
        return line(self._parent, **kwargs)

    def bar(self, **kwargs):
        return bar(self._parent, **kwargs)

    def stack(self, **kwargs):
        return stack(self._parent, **kwargs)

    def hist(self, **kwargs):
        raise NotImplementedError("Histogram plot not implemented yet!")

    def box(self, **kwargs):
        return box(self._parent, **kwargs)

    def pie(self, **kwargs):
        return pie(self._parent, **kwargs)

    def scatter(self, *args, **kwargs):
        return scatter(self._parent, *args, **kwargs)

    def sankey(self, *args, **kwargs):
        return sankey(self._parent, *args, **kwargs)


def reset_default_props(**kwargs):
    """Reset properties to initial cycle point"""
    global _DEFAULT_PROPS
    pcycle = plt.rcParams["axes.prop_cycle"]
    _DEFAULT_PROPS = {
        "color": (
            itertools.cycle(_get_standard_colors(**kwargs))
            if len(kwargs) > 0
            else itertools.cycle([x["color"] for x in pcycle])
        ),
        "marker": itertools.cycle(["o", "x", ".", "+", "*"]),
        "linestyle": itertools.cycle(["-", "--", "-.", ":"]),
    }


def default_props(reset=False, **kwargs):
    """Return current default properties

    Parameters
    ----------
    reset : bool
            if True, reset properties and return
    """
    global _DEFAULT_PROPS
    if _DEFAULT_PROPS is None or reset:
        reset_default_props(**kwargs)
    return _DEFAULT_PROPS


def mpl_args_to_meta_cols(df, **kwargs):
    """Return the kwargs values (not keys) matching a `df.meta` column name"""
    cols = set()
    for arg, value in kwargs.items():
        if is_str(value) and value in df.meta.columns:
            cols.add(value)
    return list(cols)


def assign_style_props(df, color=None, marker=None, linestyle=None, cmap=None):
    """Assign the style properties for a plot

    Parameters
    ----------
    df : pd.DataFrame
        data to be used for style properties
    """
    if color is None and cmap is not None:
        raise ValueError("`cmap` must be provided with the `color` argument")

    # determine color, marker, and linestyle for each line
    n = (
        len(df[color].unique())
        if color in df.columns
        else len(df[list(set(df.columns) & set(IAMC_IDX))].drop_duplicates())
    )
    defaults = default_props(reset=True, num_colors=n, colormap=cmap)

    props = {}
    rc = run_control()

    kinds = [("color", color), ("marker", marker), ("linestyle", linestyle)]

    for kind, var in kinds:
        rc_has_kind = kind in rc
        if var in df.columns:
            rc_has_var = rc_has_kind and var in rc[kind]
            props_for_kind = {}

            for val in df[var].unique():
                if rc_has_var and val in rc[kind][var]:
                    props_for_kind[val] = rc[kind][var][val]
                    # cycle any way to keep defaults the same
                    next(defaults[kind])
                else:
                    props_for_kind[val] = next(defaults[kind])
            props[kind] = props_for_kind

    # update for special properties only if they exist in props
    if "color" in props:
        d = props["color"]
        values = list(d.values())
        # find if any colors in our properties corresponds with special colors
        # we know about
        overlap_idx = np.isin(values, list(PYAM_COLORS.keys()))
        if overlap_idx.any():  # some exist in our special set
            keys = np.array(list(d.keys()))[overlap_idx]
            values = np.array(values)[overlap_idx]
            # translate each from pyam name, like AR6-SSP2-45 to proper color
            # designation
            for k, v in zip(keys, values):
                d[k] = PYAM_COLORS[v]
            # replace props with updated dict without special colors
            props["color"] = d
    return props


def reshape_mpl(df, x, y, idx_cols, **kwargs):
    """Reshape data from long form to "bar plot form".

    Matplotlib requires x values as the index with one column for bar grouping.
    Table values come from y values.
    """
    idx_cols = to_list(idx_cols)
    if x not in idx_cols:
        idx_cols += [x]

    # check for duplicates
    rows = df[idx_cols].duplicated()
    if any(rows):
        raise_data_error("Duplicates in plot data", df.loc[rows, idx_cols])

    # reshape the data
    df = df.set_index(idx_cols)[y].unstack(x).T

    # reindex to get correct order
    for key, value in kwargs.items():
        level = None
        if df.columns.name == key:  # single-dimension index
            axis, _values = "columns", df.columns.values
        elif df.index.name == key:  # single-dimension index
            axis, _values = "index", list(df.index)
        elif key in df.columns.names:  # several dimensions -> pd.MultiIndex
            axis, _values = "columns", get_index_levels(df.columns, key)
            level = key
        else:
            raise ValueError(f"No dimension {key} in the data!")

        # if not given, determine order based on run control (if possible)
        if value is None and key in run_control()["order"]:
            # select relevant items from run control, then add other cols
            value = [i for i in run_control()["order"][key] if i in _values]
            value += [i for i in _values if i not in value]
        df = df.reindex(**{axis: value, "level": level})

    return df


def time_col_or_year(df):
    """Return the time-col (if `df` is an IamDataFrame) or 'year'"""
    try:
        return df.time_col
    except AttributeError:
        return "year"


def pie(
    df,
    value="value",
    category="variable",
    legend=False,
    title=None,
    ax=None,
    cmap=None,
    **kwargs,
):
    """Plot data as a pie chart.

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`, :class:`pandas.DataFrame`
        Data to be plotted
    value : string, optional
        The column to use for data values
    category : string, optional
        The column to use for labels
    legend : bool, optional
        Include a legend.
    title : string, optional
        Text to use for the title.
    ax : :class:`matplotlib.axes.Axes`, optional
    cmap : string, optional
        The name of a registered colormap.
    **kwargs
        Additional arguments passed to :meth:`pandas.DataFrame.plot`.

    Returns
    -------
    ax : :class:`matplotlib.axes.Axes`
        Modified `ax` or new instance
    """

    # cast to DataFrame if necessary
    # TODO: select only relevant meta columns
    if not isinstance(df, pd.DataFrame):
        df = df.as_pandas()

    for col in set(SORT_IDX) - {category}:
        if len(df[col].unique()) > 1:
            msg = (
                "Can not plot multiple {}s in a pie plot with value={} and category={}"
            )
            raise ValueError(msg.format(col, value, category))

    if ax is None:
        fig, ax = plt.subplots()

    # get data, set negative values to explode
    _df = df.groupby(category)[value].sum()
    where = _df > 0
    explode = tuple(0 if _ else 0.2 for _ in where)
    _df = _df.abs()

    # explicitly get colors
    defaults = default_props(reset=True, num_colors=len(_df.index), colormap=cmap)[
        "color"
    ]
    rc = run_control()

    if "colors" in kwargs:
        colors = kwargs.pop("colors")
    else:
        colors = []
        for key, c in zip(_df.index, defaults):
            if category in rc["color"] and key in rc["color"][category]:
                c = rc["color"][category][key]
            colors.append(c)

    # plot data
    _df.plot(kind="pie", colors=colors, ax=ax, explode=explode, **kwargs)

    # add legend and title
    ax.legend(loc="center left", bbox_to_anchor=(1.0, 0.5), labels=_df.index)
    if not legend:
        ax.legend_.remove()
    if title:
        ax.set_title(title)

    # remove label
    ax.set_ylabel("")

    return ax


def stack(  # noqa: C901
    df,
    x=None,
    y="value",
    stack="variable",
    order=None,
    total=None,
    legend=True,
    title=True,
    ax=None,
    cmap=None,
    **kwargs,
):
    r"""Plot a stacked area chart of timeseries data

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`, :class:`pandas.DataFrame`
        Data to be plotted
    x : string, optional
        The coordinates or column of the data points for the horizontal axis;
        defaults to the time domain (if `df` is IamDataFrame) or 'year'.
    y : string, optional
        The coordinates or column of the data points for the vertical axis.
    stack : string, optional
        The column to use for stack groupings
    order : list, optional
         The order to plot the stack levels and the legend. If not specified,
         order by :meth:`run_control()['order'][\<stack\>] <pyam.run_control>`
         (where available) or alphabetical.
    total : bool or dict, optional
        If True, plot a total line with default |pyam| settings. If a dict,
        then plot the total line using the dict key-value pairs as keyword
        arguments to :meth:`matplotlib.axes.Axes.plot`.
        If None, do not plot the total line.
    legend : bool, optional
        Include a legend.
    title : bool or string, optional
        Text to use for the title, display a default if True.
    ax : :class:`matplotlib.axes.Axes`, optional
    cmap : string, optional
        The name of a registered colormap.
    **kwargs
        Additional arguments passed to :meth:`pandas.DataFrame.plot`

    Returns
    -------
    ax : :class:`matplotlib.axes.Axes`
        Modified `ax` or new instance
    """

    # default x-axis to time-col attribute from an IamDataFrame, else use "year"
    x = x or time_col_or_year(df)

    # cast to DataFrame if necessary
    # TODO: select only relevant meta columns
    if not isinstance(df, pd.DataFrame):
        df = df.as_pandas()

    for col in set(SORT_IDX) - {x, stack}:
        if len(df[col].unique()) > 1:
            msg = "Can not plot multiple {}s in stack_plot with x={}, stack={}"
            raise ValueError(msg.format(col, x, stack))

    if ax is None:
        fig, ax = plt.subplots()

    # long form to one column per stack group
    _df = reshape_mpl(df, x, y, stack, **{stack: order})

    # cannot plot timeseries that do not extend for the entire range
    has_na = _df.iloc[[0, -1]].isna().any()
    if any(has_na):
        msg = "Can not plot data that does not extend for the entire {} range"
        raise ValueError(msg.format(x))

    def as_series(index, name):
        _idx = [i[0] for i in index]
        return pd.Series([0] * len(index), index=_idx, name=name, dtype="int")

    # determine all time-indices where a timeseries crosses 0 and add to data
    _rows = pd.concat(
        [as_series(cross_threshold(_df[c], return_type=float), c) for c in _df.columns],
        axis=1,
    )
    _df = (
        pd.concat([_df, _rows.loc[_rows.index.difference(_df.index)]])
        .sort_index()
        .interpolate(method="index")
    )

    # explicitly get colors
    defaults = default_props(reset=True, num_colors=len(_df.columns), colormap=cmap)[
        "color"
    ]
    rc = run_control()
    colors = {}
    for key in _df.columns:
        c = next(defaults)
        c_in_rc = "color" in rc
        if c_in_rc and stack in rc["color"] and key in rc["color"][stack]:
            c = rc["color"][stack][key]
        colors[key] = c

    # determine positive and negative parts of the timeseries data
    _df_pos = _df.map(lambda x: max(x, 0))
    _df_neg = _df.map(lambda x: min(x, 0))

    lower = [0] * len(_df_pos)
    for col in reversed(_df_pos.columns):
        upper = _df_pos[col].fillna(0) + lower
        ax.fill_between(
            _df_pos.index,
            upper,
            lower,
            label=None,
            color=colors[col],
            linewidth=0,
            **kwargs,
        )
        lower = upper

    upper = [0] * len(_df_neg)
    for col in _df_neg.columns:
        lower = _df_neg[col].fillna(0) + upper
        # add label only on negative to have it in right order
        ax.fill_between(
            _df_neg.index,
            upper,
            lower,
            label=col,
            color=colors[col],
            linewidth=0,
            **kwargs,
        )
        upper = lower

    # add total
    if (total is not None) and total:  # cover case where total=False
        if isinstance(total, bool):  # can now assume total=True
            total = {}
        total.setdefault("label", "Total")
        total.setdefault("color", "black")
        total.setdefault("lw", 4.0)
        ax.plot(_df.index, _df.sum(axis=1), **total)

    # add legend
    ax.legend(loc="center left", bbox_to_anchor=(1.0, 0.5))
    if not legend:
        ax.legend_.remove()

    # add default labels if possible
    ax.set_xlabel(x.capitalize())
    units = df["unit"].unique()
    if len(units) == 1:
        ax.set_ylabel(units[0])

    # build a default title if possible
    _title = []
    for var in ["model", "scenario", "region", "variable"]:
        values = df[var].unique()
        if len(values) == 1:
            _title.append(f"{var}: {values[0]}")
    if title and _title:
        title = " ".join(_title) if title is True else title
        ax.set_title(title)

    return ax


def bar(  # noqa: C901
    df,
    x=None,
    y="value",
    bars="variable",
    order=None,
    bars_order=None,
    orient="v",
    legend=True,
    title=True,
    ax=None,
    cmap=None,
    **kwargs,
):
    r"""Plot data as a stacked or grouped bar chart

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`, :class:`pandas.DataFrame`
        Data to be plotted
    x : string, optional
        The coordinates or column of the data points for the horizontal axis;
        defaults to the time domain (if `df` is IamDataFrame) or 'year'.
    y : string, optional
        The coordinates or column of the data points for the vertical axis.
    bars : string, optional
        The column to use for bar groupings
    order, bars_order : list, optional
         The order to plot the levels on the x-axis and the bars (and legend).
         If not specified, order
         by :meth:`run_control()['order'][\<stack\>] <pyam.run_control>`
         (where available) or alphabetical.
    orient : string, optional
        Vertical or horizontal orientation.
    legend : bool, optional
        Include a legend.
    title : bool or string, optional
        Text to use for the title, display a default if True.
    ax : :class:`matplotlib.axes.Axes`, optional
    cmap : string, optional
        The name of a registered colormap.
    **kwargs
        Additional arguments passed to :meth:`pandas.DataFrame.plot`

    Returns
    -------
    ax : :class:`matplotlib.axes.Axes`
        Modified `ax` or new instance
    """

    # default x-axis to time-col attribute from an IamDataFrame, else use "year"
    x = x or time_col_or_year(df)

    # cast to DataFrame if necessary
    # TODO: select only relevant meta columns
    if not isinstance(df, pd.DataFrame):
        df = df.as_pandas()

    for col in set(SORT_IDX) - {x, bars}:
        if len(df[col].unique()) > 1:
            msg = "Can not plot multiple {}s in bar plot with x={}, bars={}"
            raise ValueError(msg.format(col, x, bars))

    if ax is None:
        fig, ax = plt.subplots()

    # long form to one column per bar group
    _df = reshape_mpl(df, x, y, bars, **{x: order, bars: bars_order})

    # explicitly get colors
    defaults = default_props(reset=True, num_colors=len(_df.columns), colormap=cmap)[
        "color"
    ]
    rc = run_control()
    color = []
    for key in _df.columns:
        c = next(defaults)
        if "color" in rc and bars in rc["color"] and key in rc["color"][bars]:
            c = rc["color"][bars][key]
        color.append(c)

    # change year to str to prevent pandas/matplotlib from auto-ordering (#474)
    if _df.index.name == "year":
        _df.index = map(str, _df.index)

    # plot data
    kind = "bar" if orient.startswith("v") else "barh"
    _df.plot(kind=kind, color=color, ax=ax, **kwargs)

    # add legend
    ax.legend(loc="center left", bbox_to_anchor=(1.0, 0.5))
    if not legend:
        ax.legend_.remove()

    # add default labels if possible
    if orient == "v":
        ax.set_xlabel(x.capitalize())
    else:
        ax.set_ylabel(x.capitalize())
    units = df["unit"].unique()
    if len(units) == 1 and y == "value":
        if orient == "v":
            ax.set_ylabel(units[0])
        else:
            ax.set_xlabel(units[0])

    # build a default title if possible
    _title = []
    for var in ["model", "scenario", "region", "variable"]:
        values = df[var].unique()
        if len(values) == 1:
            _title.append(f"{var}: {values[0]}")
    if title and _title:
        title = " ".join(_title) if title is True else title
        ax.set_title(title)

    return ax


def box(df, y="value", x=None, by=None, legend=True, title=None, ax=None, **kwargs):
    """Plot boxplot of data using seaborn.boxplot

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`, :class:`pandas.DataFrame`
        Data to be plotted
    y : string, optional
        The column to use for y-axis values representing the distribution
        within the boxplot
    x : string, optional
        The coordinates or column of the data points for the horizontal axis;
        defaults to the time domain (if `df` is IamDataFrame) or 'year'.
    by : string, optional
        The column for grouping y-axis values at each x-axis point,
        i.e. a 3rd dimension. Data should be categorical, not a contiuous
        variable.
    legend : bool, optional
        Include a legend.
    title : bool or string, optional
        Text to use for the title, display a default if True.
    ax : :class:`matplotlib.axes.Axes`, optional
    **kwargs
        Additional arguments passed to :meth:`pandas.DataFrame.plot`.

    Returns
    -------
    ax : :class:`matplotlib.axes.Axes`
        Modified `ax` or new instance
    """

    # default x-axis to time-col attribute from an IamDataFrame, else use "year"
    x = x or time_col_or_year(df)

    # cast to DataFrame if necessary
    # TODO: select only relevant meta columns
    if not isinstance(df, pd.DataFrame):
        df = df.as_pandas()

    if by:
        rc = run_control()
        if "palette" not in kwargs and "color" in rc and by in rc["color"]:
            # TODO this only works if all categories are defined in run_control
            palette = rc["color"][by]
            df[by] = df[by].astype("category")
            df[by].cat.set_categories(list(palette), inplace=True)
            kwargs["palette"] = palette
        else:
            df.sort_values(by, inplace=True)

    if ax is None:
        fig, ax = plt.subplots()

    # Create the plot
    sns.boxplot(x=x, y=y, hue=by, data=df, ax=ax, **kwargs)

    # Add legend
    if legend:
        ax.legend(loc=2)
        ax.legend_.set_title(
            "n=" + str(len(df[META_IDX].drop_duplicates())),
        )

    # Axes labels
    if y == "value":
        ax.set_ylabel(df.unit.unique()[0])
    else:
        ax.set_ylabel(y)

    if title:
        ax.set_title(title)

    return ax


def _get_boxes(ax, xoffset=0.05, width_weight=0.1):
    xys = {}
    widths = {}
    heights = defaultdict(list)
    for b in ax.get_children():
        if isinstance(b, mpatches.Rectangle) and b.xy != (0, 0):
            x, y = b.xy
            heights[x].append(b.get_height())
            widths[x] = b.get_width() * width_weight
            xys[x] = ((x + b.get_width()) + xoffset, 0)
    return {x: (xys[x], widths[x], sum(heights[x])) for x in xys.keys()}


def add_net_values_to_bar_plot(axs, color="k"):
    """Add net values next to an existing vertical stacked bar chart

    Parameters
    ----------
    axs : matplotlib.Axes or list thereof
    color : str, optional, default: black
        the color of the bars to add
    """
    axs = axs if isinstance(axs, Iterable) else [axs]
    for ax in axs:
        box_args = _get_boxes(ax)
        for x, args in box_args.items():
            rect = mpatches.Rectangle(*args, color=color)
            ax.add_patch(rect)


def scatter(  # noqa: C901
    df,
    x,
    y,
    legend=None,
    title=None,
    color=None,
    marker="o",
    linestyle=None,
    groupby=["model", "scenario"],
    with_lines=False,
    ax=None,
    cmap=None,
    **kwargs,
):
    """Plot data as a scatter chart.

    Parameters
    ----------
    df : class:`pyam.IamDataFrame`
        Data to be plotted
    x : str
        The coordinates or columns of the data points for the horizontal axis.
    y : str
        The coordinates or columns of the data points for the vertical axis.
    legend : bool, optional
        Include a legend. By default, show legend only if less than 13 entries.
        If a dictionary is provided, it will be used as keyword arguments
        in creating the legend.
    title : bool or string, optional
        Text to use for the title, display a default if True.
    color : string, optional
        A valid matplotlib color or column name. If a column name, common
        values will be provided the same color.
    marker : string
        A valid matplotlib marker or column name. If a column name, common
        values will be provided the same marker.
    linestyle : string, optional
        A valid matplotlib linestyle or column name. If a column name, common
        values will be provided the same linestyle.
        default: None
    groupby : list-like, optional
        Data grouping for plotting.
    with_lines : bool, optional
        Make the scatter plot with lines connecting common data.
    ax : :class:`matplotlib.axes.Axes`, optional
    cmap : string, optional
        The name of a registered colormap.
    **kwargs
        Additional arguments passed to :meth:`pandas.DataFrame.plot`.

    Returns
    -------
    ax : :class:`matplotlib.axes.Axes`
        Modified `ax` or new instance
    """

    # process the data
    xisvar = x in df.variable
    yisvar = y in df.variable

    meta_col_args = dict(color=color, marker=marker, linestyle=linestyle)
    meta_cols = mpl_args_to_meta_cols(df, **meta_col_args)

    if not xisvar and not yisvar:
        cols = [x, y] + meta_cols
        data = df.meta[cols].reset_index()
    elif xisvar and yisvar:
        # filter pivot both and rename
        dfx = (
            df.filter(variable=x)
            .as_pandas(meta_cols=meta_cols)
            .rename(columns={"value": x, "unit": "xunit"})
            .set_index(YEAR_IDX)
            .drop("variable", axis=1)
        )
        dfy = (
            df.filter(variable=y)
            .as_pandas(meta_cols=meta_cols)
            .rename(columns={"value": y, "unit": "yunit"})
            .set_index(YEAR_IDX)
            .drop("variable", axis=1)
        )
        data = dfx.join(dfy, lsuffix="_left", rsuffix="").reset_index()
    else:
        # filter, merge with meta, and rename value column to match var
        var = x if xisvar else y
        data = (
            df.filter(variable=var)
            .as_pandas(meta_cols=mpl_args_to_meta_cols(df, **kwargs))
            .rename(columns={"value": var})
        )

    # drop nan
    data.dropna(inplace=True)

    # create a plotting axes (if not given as kwarg)
    if ax is None:
        fig, ax = plt.subplots()

    # assign styling properties
    props = assign_style_props(data, **meta_col_args, cmap=cmap)

    # group data
    groups = data.dropna().groupby(groupby)

    # loop over grouped dataframe, plot data
    legend_data = []
    for name, group in groups:
        pargs = {}
        labels = []
        for key, kind, var in [
            ("c", "color", color),
            ("marker", "marker", marker),
            ("linestyle", "linestyle", linestyle),
        ]:
            if kind in props:
                label = group[var].values[0]
                pargs[key] = props[kind][group[var].values[0]]
                labels.append(repr(label).lstrip("u'").strip("'"))
            else:
                pargs[key] = var

        if len(labels) > 0:
            legend_data.append(" ".join(labels))
        else:
            legend_data.append(" ".join(name))
        kwargs.update(pargs)
        label = " ".join(group[g].iloc[0] for g in groupby)
        if with_lines:
            ax.plot(group[x], group[y], label=label, **kwargs)
        else:
            kwargs.pop("linestyle")  # scatter() can't take a linestyle
            ax.scatter(group[x], group[y], label=label, **kwargs)

    # build legend handles and labels
    handles, labels = ax.get_legend_handles_labels()
    if legend_data != [""] * len(legend_data):
        labels = sorted(list(set(tuple(legend_data))))
        idxs = [legend_data.index(d) for d in labels]
        handles = [handles[i] for i in idxs]
    if legend is not False:
        _add_legend(ax, handles, labels, legend)

    # add labels and title
    ax.set_xlabel(x)
    ax.set_ylabel(y)
    if title:
        ax.set_title(title)

    return ax


def line(  # noqa: C901
    df,
    x=None,
    y="value",
    order=None,
    legend=None,
    title=True,
    color=None,
    marker=None,
    linestyle=None,
    fill_between=None,
    final_ranges=None,
    rm_legend_label=[],
    ax=None,
    cmap=None,
    **kwargs,
):
    r"""Plot data as lines with or without markers.

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`, :class:`pandas.DataFrame`
        Data to be plotted
    x : string, optional
        The coordinates or column of the data points for the horizontal axis;
        defaults to the time domain (if `df` is IamDataFrame) or 'year'.
    y : string, optional
        The column to use as y-axis
    order : dict or list, optional
         The order of lines and the legend as :code:`{<column>: [<order>]}` or
         a list of columns where ordering should be applied. If not specified,
         order by :meth:`run_control()['order'][\<column\>] <pyam.run_control>`
         (where available) or alphabetical.
    legend : bool or dictionary, optional
        Include a legend. By default, show legend only if less than 13 entries.
        If a dictionary is provided, it will be used as keyword arguments
        in creating the legend.
    title : bool or string, optional
        Text to use for the title, display a default if True.
    color : string, optional
        A valid matplotlib color or column name. If a column name, common
        values will be provided the same color.
    marker : string, optional
        A valid matplotlib marker or column name. If a column name, common
        values will be provided the same marker.
    linestyle : string, optional
        A valid matplotlib linestyle or column name. If a column name, common
        values will be provided the same linestyle.
    fill_between : boolean or dict, optional
        Fill lines between minima/maxima of the 'color' argument. This can only
        be used if also providing a 'color' argument. If this is True, then
        default arguments will be provided to `ax.fill_between()`. If this is a
        dictionary, those arguments will be provided instead of defaults.
    final_ranges : boolean or dict, optional
        Add vertical line between minima/maxima of the 'color' argument in the
        last period plotted.  This can only be used if also providing a 'color'
        argument. If this is True, then default arguments will be provided to
        `ax.axvline()`. If this is a dictionary, those arguments will be
        provided instead of defaults.
    rm_legend_label : string or list, optional
        Remove the color, marker, or linestyle label in the legend.
    ax : :class:`matplotlib.axes.Axes`, optional
    cmap : string, optional
        The name of a registered colormap.
    **kwargs
        Additional arguments passed to :meth:`pandas.DataFrame.plot`.

    Returns
    -------
    ax : :class:`matplotlib.axes.Axes`
        Modified `ax` or new instance
    """

    # default x-axis to time-col attribute from an IamDataFrame, else use "year"
    x = x or time_col_or_year(df)

    # cast to DataFrame if necessary
    if not isinstance(df, pd.DataFrame):
        meta_col_args = dict(color=color, marker=marker, linestyle=linestyle)
        meta_cols = mpl_args_to_meta_cols(df, **meta_col_args)
        df = df.as_pandas(meta_cols=meta_cols)
        # replace np.nan by string "n/a" for proper styling
        for col in meta_cols:
            df[col] = df[col].fillna("n/a")

    # pivot data if asked for explicit variable name
    variables = df["variable"].unique()
    if x in variables or y in variables:
        keep_vars = {x, y} & set(variables)
        df = df[df["variable"].isin(keep_vars)]
        idx = list(set(df.columns) - {"value"})
        df = (
            df.reset_index()
            .set_index(idx)
            .value.unstack(level="variable")  # df -> series  # keep_vars are columns
            .rename_axis(None, axis=1)  # rm column index name
            .reset_index()
            .set_index(META_IDX)
        )
        if x != "year" and y != "year":
            df = df.drop("year", axis=1)  # years causes nan's

    if ax is None:
        fig, ax = plt.subplots()

    # assign styling properties
    props = assign_style_props(
        df, color=color, marker=marker, linestyle=linestyle, cmap=cmap
    )

    if fill_between and "color" not in props:
        raise ValueError("Must use `color` kwarg if using `fill_between`")
    if final_ranges and "color" not in props:
        raise ValueError("Must use `color` kwarg if using `final_ranges`")

    # prepare a dict for ordering, reshape data for use in line_plot
    idx_cols = list(df.columns.drop(y))
    if not isinstance(order, dict):
        order = {i: None for i in order or idx_cols}
    df = reshape_mpl(df, x, y, idx_cols, **order)

    # determine the columns that should go into the legend
    idx_cols.remove(x)
    title_cols = []
    y_label = None
    for col in idx_cols:
        values = get_index_levels(df.columns, col)
        if len(values) == 1 and col not in [color, marker, linestyle]:
            if col == "unit" and y == "value":
                y_label = values[0]
            elif col == y and col != "value":
                y_label = values[0]
            else:
                if col != "unit":
                    title_cols.append(f"{col}: {values[0]}")
            if isinstance(df.columns, pd.MultiIndex):
                df.columns = df.columns.droplevel(col)
            else:  # cannot drop last remaining level, replace by empty list
                df.columns = [""]

    # determine index of column name in reshaped dataframe
    prop_idx = {}
    for kind, var in [("color", color), ("marker", marker), ("linestyle", linestyle)]:
        if var is not None and var in df.columns.names:
            prop_idx[kind] = df.columns.names.index(var)

    # pop label to avoid multiple values for plot-kwarg
    label = kwargs.pop("label", None)

    # plot data, keeping track of which legend labels to apply
    for col, data in df.items():
        # handle case where columns are not strings or only have 1 dimension
        col = list(map(str, to_list(col)))
        pargs = {}
        labels = []
        # build plotting args and line legend labels
        for key, kind, var in [
            ("c", "color", color),
            ("marker", "marker", marker),
            ("linestyle", "linestyle", linestyle),
        ]:
            if kind in props:
                _label = col[prop_idx[kind]]
                pargs[key] = props[kind][_label]
                if kind not in to_list(rm_legend_label):
                    labels.append(repr(_label).lstrip("u'").strip("'"))
            else:
                pargs[key] = var
        kwargs.update(pargs)
        data = data.dropna()
        data.plot(ax=ax, label=label or " - ".join(labels if labels else col), **kwargs)

    if fill_between:
        _kwargs = {"alpha": 0.25} if fill_between in [True, None] else fill_between
        data = df.T
        columns = data.columns
        # get outer boundary mins and maxes
        allmins = data.groupby(color).min()
        intermins = (
            data.dropna(axis=1)
            .groupby(color)
            .min()  # nonan data
            .reindex(columns=columns)  # refill with nans
            .T.interpolate(method="index")
            .T  # interpolate
        )
        mins = pd.concat([allmins, intermins]).groupby(level=0).min()
        allmaxs = data.groupby(color).max()
        intermaxs = (
            data.dropna(axis=1)
            .groupby(color)
            .max()  # nonan data
            .reindex(columns=columns)  # refill with nans
            .T.interpolate(method="index")
            .T  # interpolate
        )
        maxs = pd.concat([allmaxs, intermaxs]).groupby(level=0).max()
        # do the fill
        for idx in mins.index:
            ymin = mins.loc[idx]
            ymax = maxs.loc[idx]
            ax.fill_between(
                ymin.index, ymin, ymax, facecolor=props["color"][idx], **_kwargs
            )

    # add bars to the end of the plot showing range
    if final_ranges:
        # have to explicitly draw it to get the tick labels (these change once
        # you add the vlines)
        plt.gcf().canvas.draw()
        _kwargs = {"linewidth": 2} if final_ranges in [True, None] else final_ranges
        first = df.index[0]
        final = df.index[-1]
        mins = df.T.groupby(color).min()[final]
        maxs = df.T.groupby(color).max()[final]
        ymin, ymax = ax.get_ylim()
        ydiff = ymax - ymin
        xmin, xmax = ax.get_xlim()
        xdiff = xmax - xmin
        xticks = ax.get_xticks()
        xlabels = ax.get_xticklabels()
        # 1.5% increase seems to be ok per extra line
        extra_space = 0.015
        for i, idx in enumerate(mins.index):
            xpos = final + xdiff * extra_space * (i + 1)
            _ymin = (mins[idx] - ymin) / ydiff
            _ymax = (maxs[idx] - ymin) / ydiff
            ax.axvline(
                xpos, ymin=_ymin, ymax=_ymax, color=props["color"][idx], **_kwargs
            )
        # for equal spacing between xmin and first datapoint and xmax and last
        # line
        ax.set_xlim(xmin, xpos + first - xmin)
        ax.set_xticks(xticks)
        ax.set_xticklabels(xlabels)

    # build unique legend handles and labels
    if legend is not False:
        handles, labels = (np.array(i) for i in ax.get_legend_handles_labels())
        if label is not None:  # label given explicitly via kwarg
            _add_legend(ax, handles, labels, legend)
        else:
            _, idx = np.unique(labels, return_index=True)
            idx.sort()
            _add_legend(ax, handles[idx], labels[idx], legend)

    # add default labels if possible
    ax.set_xlabel(x.title())
    ax.set_ylabel(y_label or y.title())

    # show a default title from columns with a unique value or a custom title
    if title:
        ax.set_title(" - ".join(title_cols) if title is True else title)

    return ax


def _add_legend(ax, handles, labels, legend):
    if legend is None and len(labels) >= MAX_LEGEND_LABELS:
        logger.info(f">={MAX_LEGEND_LABELS} labels, not applying legend")
    else:
        legend = {} if legend in [True, None] else legend
        loc = legend.pop("loc", "best")
        outside = loc.split(" ")[1] if loc.startswith("outside ") else False
        _legend = OUTSIDE_LEGEND[outside] if outside else dict(loc=loc)
        _legend.update(legend)
        ax.legend(handles, labels, **_legend)


def set_panel_label(label, ax=None, x=0.05, y=0.9):
    """Add a panel label to the figure/axes, by default in the top-left corner

    Parameters
    ----------
    label : str
        text to be added as panel label
    ax : matplotlib.Axes, optional
        panel to which to add the panel label
    x : number, default 0.05
        relative location of label to x-axis
    y : number, default 0.9
        relative location of label to y-axis
    """

    def _lim_loc(lim, loc):
        return lim[0] + (lim[1] - lim[0]) * loc

    if ax is not None:
        ax.text(_lim_loc(ax.get_xlim(), x), _lim_loc(ax.get_ylim(), y), label)
    else:
        plt.text(_lim_loc(plt.xlim(), x), _lim_loc(plt.ylim(), y), label)



import logging

import pandas as pd

from pyam.index import get_index_levels

logger = logging.getLogger(__name__)

try:
    import plotly.graph_objects as go

    HAS_PLOTLY = True
except ImportError:  # pragma: no cover
    go = None
    HAS_PLOTLY = False


def sankey(df, mapping):
    """Plot a sankey diagram

    It is currently only possible to create this diagram for single years.

    Parameters
    ----------
    df : :class:`pyam.IamDataFrame`
        Data to be plotted
    mapping : dict
        Assigns the source and target component of a variable

        .. code-block:: python

            {
                variable: (source, target),
            }

        Returns
        -------
        fig : :class:`plotly.graph_objects.Figure`
    """
    if not HAS_PLOTLY:  # pragma: no cover
        raise ImportError(
            "Missing optional dependency `plotly`, use pip or conda to install"
        )
    # Check for duplicates
    for col in [name for name in df.dimensions if name != "variable"]:
        levels = get_index_levels(df._data, col)
        if len(levels) > 1:
            raise ValueError(f"Non-unique values in column {col}: {levels}")

    # Concatenate the data with source and target columns
    _df = pd.DataFrame.from_dict(
        mapping, orient="index", columns=["source", "target"]
    ).merge(df._data, how="left", left_index=True, right_on="variable")
    label_mapping = {
        label: i
        for i, label in enumerate(set(pd.concat([_df["source"], _df["target"]])))
    }
    _df.replace(label_mapping, inplace=True)
    region = get_index_levels(_df, "region")[0]
    unit = get_index_levels(_df, "unit")[0]
    year = get_index_levels(_df, "year")[0]
    fig = go.Figure(
        data=[
            go.Sankey(
                valuesuffix=unit,
                node=dict(
                    pad=15,
                    thickness=10,
                    line=dict(color="black", width=0.5),
                    label=pd.Series(list(label_mapping)),
                    hovertemplate="%{label}: %{value}<extra></extra>",
                    color="blue",
                ),
                link=dict(
                    source=_df.source,
                    target=_df.target,
                    value=_df.value,
                    hovertemplate='"%{source.label}" to "%{target.label}": \
                %{value}<extra></extra>',
                ),
            )
        ]
    )
    fig.update_layout(title_text=f"region: {region}, year: {year}", font_size=10)
    return fig



import re

import numpy as np

from pyam import IamDataFrame
from pyam.str import is_str
from pyam.utils import pattern_match, to_list

# columns from UNFCCC data that can be used for variable names
NAME_COLS = ["category", "classification", "measure", "gas"]

# UNFCCC-reader instance (instantiated at first use)
_READER = None


def read_unfccc(
    party_code,
    gases=None,
    tier=None,
    mapping=None,
    model="UNFCCC",
    scenario="Data Inventory",
):
    """Read data from the UNFCCC Data Inventory

    This function is a wrappter for the
    :meth:`unfccc_di_api.UNFCCCApiReader.query`.

    The data returned from the UNFCCC Data Inventory is transformed
    into a structure similar to the format used in IPCC reports and
    IAM model comparison projects. For compatibility with the
    `iam-units <https://github.com/IAMconsortium/units>`_ package
    and the :meth:`convert_unit <IamDataFrame.convert_unit>`,
    emissions species are formatted to standard text ('CO2')
    instead of subscripts ('CO₂') and the unit 'CO₂ equivalent'
    used by UNFCCC is replaced by 'CO2e'.

    Parameters
    ----------
    party_code : str
        ISO3-style code for UNFCCC party (country)
    gases : str or list of str, optional
        Emission species to be queried from the data inventory can be stated
        as subscript-format ('CO₂') or simple text ('CO2')
    tier : int or list of int
        Pre-specified groupings of UNFCCC data to a variable naming format
        used in IPCC reports and IAM model comparison projects
    mapping : dict, optional
        Mapping to cast UNFCCC-data columns into IAMC-style variables, e.g.

        .. code-block:: python

          {
              "Emissions|{gas}|Energy": ("1.  Energy", "*", "*", "*"),
          }

        where the tuple corresponds to filters for the columns
        ``["category", "classification", "measure", "gas"]``
        and ``{<col>}`` tags in the key are replaced by the column value.
    model : str, optional
        Name to be used as model identifier
    scenario : str, optional
        Name to be used as scenario identifier

    Returns
    -------
    :class:`IamDataFrame`

    Notes
    -----
    This method is currently not tested due to a change in the UNFCCC-DI API,
    which sometimes causes a `JsonDecodeError`.
    See https://github.com/pik-primap/unfccc_di_api/issues/74 for more info.
    """

    # import packages for functions with low-frequency usage only when needed
    import unfccc_di_api

    # check that only one of `tier` or `mapping` is provided
    if (tier is None and mapping is None) or (tier is not None and mapping is not None):
        raise ValueError("Please specify either `tier` or `mapping`.")

    global _READER
    if _READER is None:
        _READER = unfccc_di_api.UNFCCCApiReader()

    # retrieve data, drop non-numeric data and base year
    data = _READER.query(party_code=party_code, gases=to_list(gases))
    data = data[~np.isnan(data.numberValue)]
    data = data[data.year != "Base year"]

    # create the mapping from the data if `tier` is given
    if tier is not None:
        _category = data.category.unique()
        mapping = {}

        for t in to_list(tier):
            # treatment of tear 1
            if t == 1:
                pattern = re.compile(".\\.  ")  # pattern of top-level category
                for i in [i for i in _category if pattern.match(i)]:
                    key = "Emissions|{gas}|" + i[4:]
                    mapping[key] = (
                        i,
                        "Total for category",
                        "Net emissions/removals",
                        "*",
                    )
            else:
                raise ValueError(f"Unknown value for `tier`: {t}")

    # add new `variable` column, iterate over mapping to determine variables
    data["variable"] = None
    for variable, value in mapping.items():
        matches = np.array([True] * len(data))
        for i, col in enumerate(NAME_COLS):
            matches &= pattern_match(data[col], value[i])

        data.loc[matches, "variable"] = data.loc[matches].apply(
            _compile_variable, variable=variable, axis=1
        )

    # drop unspecified rows and columns, rename value column
    cols = ["party", "variable", "unit", "year", "gas", "numberValue"]
    data = data.loc[[is_str(i) for i in data.variable], cols]
    data.rename(columns={"numberValue": "value"}, inplace=True)

    # append `gas` to unit, drop `gas` column
    data.loc[:, "unit"] = data.apply(_compile_unit, axis=1)
    data.drop(columns="gas", inplace=True)

    return IamDataFrame(data, model=model, scenario=scenario, region="party")


def _compile_variable(i, variable):
    """Translate UNFCCC columns into an IAMC-style variable"""
    if i["variable"]:
        raise ValueError("Conflict in variable mapping.")
    return variable.format(**{c: i[c] for c in NAME_COLS})


def _compile_unit(i):
    """Append gas to unit and update CO2e for pint/iam-unit compatibility"""
    if " equivalent" in i["unit"]:
        return i["unit"].replace("CO2 equivalent", "CO2e")
    if i["unit"] in ["kt", "t"]:
        return " ".join([i["unit"], i["gas"]])
    else:
        return i["unit"]



import logging

import pandas as pd

from pyam.utils import format_n, make_index

logger = logging.getLogger(__name__)


def _validate(df, upper_bound, lower_bound, exclude_on_fail, **kwargs):  # noqa: C901
    _df = df._data[df.slice(**kwargs)]
    if _df.empty:
        logger.warning("No data matches filters, skipping validation.")
        return

    failed_validation = []
    if upper_bound is not None:
        failed_validation.append(_df[_df > upper_bound])
    if lower_bound is not None:
        failed_validation.append(_df[_df < lower_bound])
    if not failed_validation:
        return
    _df = pd.concat(failed_validation).sort_index()

    if not _df.empty:
        msg = "{} of {} data points do not satisfy the criteria"
        logger.warning(msg.format(len(_df), len(df.data)))

        if exclude_on_fail and len(_df) > 0:
            _exclude_on_fail(df, _df)
        return _df.reset_index()


def _check_rows(rows, check, in_range=True, return_test="any"):
    """Check all rows to be in/out of a certain range and provide testing on
    return values based on provided conditions

    Parameters
    ----------
    rows : pd.DataFrame
        data rows
    check : dict
        dictionary with possible values of 'up', 'lo', and 'year'
    in_range : bool, optional
        check if values are inside or outside of provided range
    return_test : str, optional
        possible values:
            - 'any': default, return scenarios where check passes for any entry
            - 'all': test if all values match checks, if not, return empty set
    """
    valid_checks = {"up", "lo", "year"}
    if not set(check.keys()).issubset(valid_checks):
        msg = "Unknown checking type: {}"
        raise ValueError(msg.format(check.keys() - valid_checks))
    if "year" not in check:
        where_idx = set(rows.index)
    else:
        if "time" in rows.index.names:
            _years = rows.index.get_level_values("time").year
        else:
            _years = rows.index.get_level_values("year")
        where_idx = set(rows.index[_years == check["year"]])
        rows = rows.loc[list(where_idx)]

    up_op = rows.values.__le__ if in_range else rows.values.__gt__
    lo_op = rows.values.__ge__ if in_range else rows.values.__lt__

    check_idx = []
    for bd, op in [("up", up_op), ("lo", lo_op)]:
        if bd in check:
            check_idx.append(set(rows.index[op(check[bd])]))

    if return_test == "any":
        ret = where_idx & set.union(*check_idx)
    elif return_test == "all":
        ret = where_idx if where_idx == set.intersection(*check_idx) else set()
    else:
        raise ValueError(f"Unknown return test: {return_test}")
    return ret


def _exclude_on_fail(df, index):
    """Assign a selection of scenarios as `exclude: True`"""

    if not isinstance(index, pd.MultiIndex):
        index = make_index(index, cols=df.index.names)

    df.exclude[index] = True
    n = len(index)
    logger.info(
        format_n(n, "scenario")
        + " failed validation and will be set as `exclude=True`."
    )



import pandas as pd

from pyam import IamDataFrame


def read_worldbank(model="World Bank", scenario="WDI", **kwargs) -> IamDataFrame:
    """Read data from the World Bank Data Catalogue and return as IamDataFrame

    This function is a simple wrapper for the function
    :func:`wbdata.get_dataframe`. Import the module to retrieve/search
    the list of indicators (and their id's), countries, etc.

    .. code-block:: python

        import wbdata as wb

    Parameters
    ----------
    model : str, optional
        The `model` name to be used for the returned timeseries data.
    scenario : str, optional
        The `scenario` name to be used for the returned timeseries data.
    **kwargs
        passed to :func:`wbdata.get_dataframe`

    Notes
    -----
    The function :func:`wbdata.get_dataframe` takes an `indicators`
    argument, which is a dictionary where the keys are desired indicators and the values
    are the desired column names. If the `indicators` passed to :func:`read_worldbank`
    is a single indicator code string, we should instead use :func:`wbdata.get_series`.

    The function :func:`wbdata.get_dataframe` does not return a unit,
    but it can be collected for some indicators using the function
    :func:`wbdata.get_indicators`.
    In the current implementation, unit is defined as `n/a` for all data;
    this can be enhanced later (if there is interest from users).

    Returns
    -------
    :class:`IamDataFrame`
    """
    # import packages for functions with low-frequency usage only when needed
    # also, there seems to be an issue with wbdata on Mac OS
    # see https://github.com/OliverSherouse/wbdata/issues/74
    import wbdata  # noqa: F401

    data: pd.DataFrame = wbdata.get_dataframe(**kwargs)
    value = data.columns
    data.reset_index(inplace=True)
    data.rename(columns={"date": "year"}, inplace=True)
    df = IamDataFrame(
        data,
        model=model,
        scenario=scenario,
        value=value,
        unit="n/a",
        region="country",
    )
    # TODO use wb.get_indicators to retrieve correct units (where available)

    # if `indicators` is a mapping, use it for renaming
    if "indicators" in kwargs and isinstance(kwargs["indicators"], dict):
        df.rename(variable=kwargs["indicators"], inplace=True)

    return df



def _compute_bias(df, name, method, axis):
    """Internal implementation for computing bias weights"""
    if method == "count":
        # invert from the count to obtain the weighting factor
        count = 1 / df.exclude.groupby(axis).count()
        count.name = name
        df.meta = df.meta.join(count, on=axis, how="outer")
    else:
        raise ValueError(f"Unknown method '{method}' for computing bias weights.")



import logging

import ixmp4
import pandas as pd
from ixmp4.core.region import RegionModel
from ixmp4.core.unit import UnitModel
from ixmp4.data.abstract import DataPoint

logger = logging.getLogger(__name__)


def read_ixmp4(
    platform: ixmp4.Platform | str,
    *,
    default_only: bool = True,
    model: str | list[str] | None = None,
    scenario: str | list[str] | None = None,
    region: str | list[str] | None = None,
    variable: str | list[str] | None = None,
    unit: str | list[str] | None = None,
    year: int | list[int] | None = None,
):
    """Read scenario runs from an ixmp4 platform database instance

    Parameters
    ----------
    platform : :class:`ixmp4.Platform` or str
        The ixmp4 platform database instance to which the scenario data is saved.
    default_only : :class:`bool`, optional
        Read only default runs.
    model, scenario, region, variable, unit : str or list of str, optional
        Filter by these dimensions.
    year : int or list of int, optional
        Filter by time domain.
    """
    from pyam import IamDataFrame

    if not isinstance(platform, ixmp4.Platform):
        platform = ixmp4.Platform(platform)

    # TODO This may have to be revised, see https://github.com/iiasa/ixmp4/issues/72
    meta_filters = dict(
        run=dict(default_only=default_only, model=model, scenario=scenario)
    )
    iamc_filters = dict(run=dict(default_only=default_only))
    for key, value in (
        ("model", model),
        ("scenario", scenario),
        ("region", region),
        ("variable", variable),
        ("unit", unit),
        ("year", year),
    ):
        if value is not None:
            iamc_filters[key] = value

    data = platform.iamc.tabulate(**iamc_filters)
    meta = platform.meta.tabulate(**meta_filters)

    if data.empty:
        raise ValueError("No scenario data with filters " + str(iamc_filters))

    # if default-only, simplify to standard IAMC index, add `version` as meta indicator
    if default_only:
        index = ["model", "scenario"]
        meta_version = (
            data[index + ["version"]]
            .drop_duplicates()
            .rename(columns={"version": "value"})
        )
        meta_version["key"] = "version"
        meta = pd.concat([meta.drop(columns="version"), meta_version])
        data.drop(columns="version", inplace=True)
    else:
        index = ["model", "scenario", "version"]

    return IamDataFrame(data, meta=meta, index=index)


def read_run(
    run: ixmp4.Run,
    region: str | list[str] | None = None,
    variable: str | list[str] | None = None,
    unit: str | list[str] | None = None,
    year: int | list[int] | None = None,
):
    """Return timeseries data and meta indicatos from an ixmp4 run as IamDataFrame

    Parameters
    ----------
    run : :class:`ixmp4.Run`
        An ixmp4 run instance from which to read timeseries data and meta indicators
    region, variable, unit : str or list of str, optional
        Filter timeseries data by these dimensions.
    year : int or list of int, optional
        Filter timeseries data by time domain.
    """
    from pyam import IamDataFrame

    if year is not None:
        raise NotImplementedError("Filter by 'year' not implemented in ixmp4.")

    meta = pd.DataFrame.from_dict(run.meta, orient="index").T
    meta.index = pd.MultiIndex.from_tuples(
        [(run.model.name, run.scenario.name)], name=["model", "scenario"]
    )
    meta["version"] = run.version

    return IamDataFrame(
        data=run.iamc.tabulate(
            region=region,
            variable=variable,
            unit=unit,
        ),
        meta=meta,
        model=run.model.name,
        scenario=run.scenario.name,
    )


def write_to_ixmp4(platform: ixmp4.Platform | str, df, checkpoint_message: str):
    """Save all scenarios as new default runs in an ixmp4 platform database instance

    Parameters
    ----------
    platform : :class:`ixmp4.Platform` or str
        The ixmp4 platform database instance to which the scenario data is saved
    df : pyam.IamDataFrame
        The IamDataFrame instance with scenario data
    checkpoint_message : str
        The message for the ixmp4 checkpoint (similar to a commit message).
    """
    if df.extra_cols:
        raise NotImplementedError(
            "Only data with standard IAMC columns can be written to an ixmp4 platform."
        )

    if df.time_domain not in ["year", "datetime"]:
        raise NotImplementedError(
            "Only data with time domain 'year' or 'datetime' can be written to an ixmp4"
            " platform."
        )

    if not isinstance(platform, ixmp4.Platform):
        platform = ixmp4.Platform(platform)

    _validate_dimensions(platform, df)

    # The "version" meta-indicator, added when reading from an ixmp4 platform,
    # should not be written to the platform
    if "version" in df.meta.columns:
        logger.warning(
            "The `meta.version` column was dropped when writing to the ixmp4 platform."
        )
        meta = df.meta.drop(columns="version")
    else:
        meta = df.meta.copy()

    # Create runs and add IAMC timeseries data and meta indicators
    for model, scenario in df.index:
        _df = df.filter(model=model, scenario=scenario)

        run = platform.runs.create(model=model, scenario=scenario)

        with run.transact(checkpoint_message):
            if df.time_domain == "year":
                run.iamc.add(_df.data)
            elif df.time_domain == "datetime":
                run.iamc.add(
                    _df.data.rename(columns={"time": "step_datetime"}),
                    type=DataPoint.Type.DATETIME,
                )

            if not meta.empty:
                run.meta = dict(meta.loc[(model, scenario)])
            run.set_as_default()


def _validate_dimensions(platform, df):
    """Ensure that all regions and units in the DataFrame exist in the platform"""
    for dimension, values, model in [
        ("regions", df.region, RegionModel),
        ("units", df.unit, UnitModel),
    ]:
        platform_values = getattr(platform, dimension).tabulate().name.values
        if missing := set(values).difference(platform_values):
            raise model.NotFound(
                ", ".join(missing)
                + f". Use `Platform.{dimension}.create()` to add missing elements."
            )



import pandas as pd

from pyam.utils import print_list


class IamSlice(pd.Series):
    """A slice object of the IamDataFrame timeseries data index"""

    @property
    def _constructor(self):
        return IamSlice

    _internal_names = pd.Series._internal_names + ["_iamcache"]
    _internal_names_set = set(_internal_names)

    def __init__(self, data=None, index=None, **kwargs):
        super().__init__(data, index, **kwargs)
        self._iamcache = dict()

    def __dir__(self):
        return self.dimensions + super().__dir__()

    def __getattr__(self, attr):
        try:
            return super().__getattr__(attr)
        except AttributeError:
            cache = object.__getattribute__(self, "_iamcache")
            ret = cache.get(attr)
            if ret is not None:
                return ret.tolist()

            if attr in self.dimensions:
                ret = cache[attr] = self.index[self].unique(level=attr)
                return ret.tolist()

            raise

    def __len__(self):
        return self.sum()

    @property
    def dimensions(self):
        """Return the list of index names & data coordinates"""
        return self.index.names

    @property
    def time(self):
        """The time index, i.e., axis labels related to the time domain.

        Returns
        -------
        - A :class:`pandas.Index` (dtype 'int64') if the :attr:`time_domain` is 'year'
        - A :class:`pandas.DatetimeIndex` if the time-domain is 'datetime'
        - A :class:`pandas.Index` if the time-domain is 'mixed'
        """
        ret = self._iamcache.get("time")
        if ret is None:
            ret = self._iamcache["time"] = (
                self.index[self].unique(level=self.time_col).rename("time")
            )
        return ret

    @property
    def time_col(self):
        return "year" if "year" in self.dimensions else "time"

    def __repr__(self):
        return self.info()

    def info(self, n=80):
        """Print a summary of the represented index dimensions and data coordinates

        Parameters
        ----------
        n : int
            The maximum line length
        """
        # concatenate list of index dimensions and levels
        info = f"{type(self)}\nIndex dimensions and data coordinates:\n"
        c1 = max([len(i) for i in self.dimensions]) + 1
        c2 = n - c1 - 5
        info += "\n".join(
            [
                f"   {i:{c1}}: {print_list(getattr(self, i), c2)}"
                for i in self.dimensions
            ]
        )

        return info



import logging
import re

import iam_units
import pint

from pyam.index import replace_index_values

logger = logging.getLogger(__name__)


def convert_unit(
    df, current, to, factor=None, registry=None, context=None, inplace=False
):
    """Internal implementation of unit conversion with explicit kwargs"""
    ret = df.copy() if not inplace else df

    # Mask for rows having *current* units to be converted, args for replace
    try:
        where = ret._data.index.get_loc_level(current, "unit")[0]
    except KeyError:
        return None if inplace else ret

    index_args = [ret._data, "unit", {current: to}]

    if factor:
        # Short code path: use an explicit conversion factor, don't use pint
        ret._data.loc[where] *= factor
        ret._data.index = replace_index_values(*index_args)
        return None if inplace else ret

    # Convert using a pint.UnitRegistry; default the one from iam_units
    registry = registry or iam_units.registry

    # Make versions without -equiv
    _current, _to = (i.replace("-equiv", "") for i in [current, to])
    # Pair of (magnitude, unit)
    qty = [ret._data.loc[where].values, _current]

    try:
        # Create a vector pint.Quantity and convert it ordinarily
        result = registry.Quantity(*qty).to(
            _to, context if context is not None else pint.Context()
        )
    except pint.UndefinedUnitError:
        # *qty* might include a GHG species; try GWP conversion
        result, _ = convert_gwp(context, qty, _to)
    except AttributeError:
        # .Quantity() did not exist
        raise TypeError(f"{registry} must be `pint.UnitRegistry`") from None

    # Copy values from the result Quantity and assign units
    ret._data.loc[where] = result.magnitude
    ret._data.index = replace_index_values(*index_args)

    return None if inplace else ret


# GWP conversion using iam_units

#: Supported lower-case aliases for chemical symbols of GHG species. See
#: :meth:`.convert_unit`.
# Keys and values can only differ in case; other items will have no effect.
SPECIES_ALIAS = {
    "ch4": "CH4",
    "co2": "CO2",
    "CO2-equiv": "CO2e",
    "co2_eq": "CO2_eq",
    "co2e": "CO2e",
    "co2eq": "CO2eq",
    "n2o": "N2O",
    "nh3": "NH3",
    "nox": "NOx",
}


# Thin wrapper around pint.UndefinedUnitError to provide a usage hint
class UndefinedUnitError(pint.UndefinedUnitError):
    def __str__(self):
        return super().__str__() + (
            "\nGWP conversion with IamDataFrame.convert_unit() requires a "
            "*context* and mass-based *to* units."
        )


def extract_species(expr):
    """Handle supported expressions for GHG species and units."""
    # Split *expr* into 1 or 3 strings. Unlike iam_units, re.IGNORECASE is used
    # to match e.g. lower-case 'ch4'.
    parts = re.split(
        iam_units.emissions.pattern.pattern, expr, maxsplit=1, flags=re.IGNORECASE
    )

    if len(parts) == 1:
        # No split occurred. *expr* is only a species ('CO2e') without units.
        species, units = parts[0], None
    else:
        # An expression with both units and species name ('kg CO2e / year').
        # parts[1] is the species, others are pre-/suffix ('kg ', ' / year').
        species, units = parts[1], (parts[0] + parts[2])

    # Convert allowed lower-case aliases to chemical symbols
    return SPECIES_ALIAS.get(species, species), units


def convert_gwp(metric, qty, to):
    """Helper for :meth:`convert_unit` to perform GWP conversions."""

    # Extract the species from *qty* and *to*, allowing supported aliases
    species_from, units_from = extract_species(qty[1])
    species_to, units_to = extract_species(to)

    try:
        # Convert using a (magnitude, unit) tuple with only units, and explicit
        # input and output units
        result = iam_units.convert_gwp(
            metric, (qty[0], units_from), species_from, species_to
        )
    except (AttributeError, ValueError):
        # Missing *metric*, or *species_to* contains invalid units. pyam
        # promises UndefinedUnitError in these cases. Use a subclass (above) to
        # add a usage hint.
        raise UndefinedUnitError(species_to) from None
    except pint.DimensionalityError:
        # Provide an exception with the user's inputs
        raise pint.DimensionalityError(qty[1], to) from None

    # Other exceptions are not caught and will pass up through convert_unit()

    if units_to:
        # Also convert the units
        result = result.to(units_to)
    else:
        # *to* was only a species name. Provide units based on input and the
        # output species name.
        to = iam_units.format_mass(result, species_to, spec=":~")

    return result, to



import json
import logging
import os.path
from functools import lru_cache
from io import StringIO
from pathlib import Path

import httpx
import ixmp4
import jwt
import numpy as np
import pandas as pd
import requests
import yaml
from ixmp4.cli.platforms import tabulate_manager_platforms
from ixmp4.conf import settings
from ixmp4.conf.auth import ManagerAuth
from requests.auth import AuthBase

from pyam.core import IamDataFrame
from pyam.exceptions import deprecation_warning
from pyam.ixmp4 import read_ixmp4
from pyam.str import is_str
from pyam.utils import (
    IAMC_IDX,
    META_IDX,
    is_list_like,
    pattern_match,
)

logger = logging.getLogger(__name__)
# set requests-logger to WARNING only
logging.getLogger("requests").setLevel(logging.WARNING)

_AUTH_URL = "https://api.manager.ece.iiasa.ac.at"
_CITE_MSG = """
You are connected to the {} scenario explorer hosted by IIASA.
 If you use this data in any published format, please cite the
 data as provided in the explorer guidelines: {}
""".replace("\n", "")
IXMP4_LOGIN = "Please run `ixmp4 login <username>` in a console"

# path to local configuration settings
DEFAULT_IIASA_CREDS = Path("~").expanduser() / ".local" / "pyam" / "iiasa.yaml"


def platforms() -> None:
    """Print a list of available ixmp4 platforms hosted by IIASA

    See Also
    --------
    ixmp4.conf.settings.manager.list_platforms
    """
    tabulate_manager_platforms(ixmp4.conf.settings.manager.list_platforms())


def _read_config(file):
    """Read username and password for IIASA API connection from file"""
    with open(file) as stream:
        creds = yaml.safe_load(stream)

    return ManagerAuth(**creds, url=str(settings.manager_url))


def _check_response(r, msg="Error connecting to IIASA database", error=RuntimeError):
    if not r.ok:
        raise error(f"{msg}: {r.text}")


class SceSeAuth(AuthBase):
    def __init__(self, creds: str = None, auth_url: str = _AUTH_URL):
        """Connection to the Scenario Services manager service for authentication.

        Parameters
        ----------
        creds : pathlib.Path or str, optional
            Path to a file with authentication credentials. This feature is deprecated,
            please run `ixmp4 login <username>` in a console instead.
        auth_url : str, optional
            Url of the authentication service
        """
        self.client = httpx.Client(base_url=auth_url, timeout=10.0, http2=True)

        if creds is None:
            if DEFAULT_IIASA_CREDS.exists():
                deprecation_warning(
                    f"{IXMP4_LOGIN} and manually delete the file "
                    f"'{DEFAULT_IIASA_CREDS}'. Using a pyam-credentials file",
                )
                self.auth = _read_config(DEFAULT_IIASA_CREDS)
            else:
                self.auth = ixmp4.conf.settings.default_auth
        elif isinstance(creds, Path) or is_str(creds):
            deprecation_warning(f"{IXMP4_LOGIN}.", "Using a pyam-credentials file")
            self.auth = _read_config(creds)
        else:
            raise DeprecationWarning(
                "Passing credentials as clear-text is not allowed. "
                f"{IXMP4_LOGIN} instead."
            )

        # self.auth is None if connection to manager service cannot be established
        if self.auth is None:
            raise httpx.ConnectError("No connection to IIASA manager service.")

        # explicit token for anonymous login is not necessary for ixmp4 platforms
        # but is required for legacy Scenario Explorer databases
        if self.auth.user.username == "@anonymous":
            self._get_anonymous_token()

        else:
            self.user = self.auth.user.username
            self.access_token = self.auth.access_token

    def _get_anonymous_token(self):
        r = self.client.get("/legacy/anonym/")
        if r.status_code >= 400:
            raise ValueError("Unknown API error: " + r.text)
        self.user, self.access_token = None, r.json()

    def __call__(self):
        try:
            # raises jwt.ExpiredSignatureError if token is expired
            jwt.decode(
                self.access_token,
                options={"verify_signature": False, "verify_exp": True},
            )
        except jwt.ExpiredSignatureError:
            if self.auth.user.username == "@anonymous":
                self._get_anonymous_token()
            else:
                self.auth.refresh_or_reobtain_jwt()
                self.access_token = self.auth.access_token

        return {"Authorization": "Bearer " + self.access_token}


class Connection:
    """A class to facilitate querying an IIASA Scenario Explorer database API

    Parameters
    ----------
    name : str, optional
        The name of a database API.
        Use :attr:`valid_connections <pyam.iiasa.Connection.valid_connections>`
        for a list of available APIs.
    creds : pathlib.Path or str, optional
        Path to a file with authentication credentials. This feature is deprecated,
        please run `ixmp4 login <username>` in a console instead.
    auth_url : str, optional
        custom authentication server URL

    Notes
    -----
    Credentials (username & password) are not required to access public |ixmp4|
    or Scenario Explorer databases (i.e., with Guest login).
    """

    def __init__(self, name=None, creds=None, auth_url=_AUTH_URL):
        self._auth_url = auth_url  # scenario services manager API
        self._base_url = None  # database connection API

        self.auth = SceSeAuth(creds=creds, auth_url=self._auth_url)

        # connect if provided a name
        self._connected = None
        if name:
            self.connect(name)

        if self.auth.user is not None:
            logger.info(f"You are connected as user '{self.auth.user}'")
        else:
            logger.info("You are connected as an anonymous user")

    @property
    @lru_cache
    def _connection_map(self):
        # TODO: application-list will be reimplemented in conjunction with ixmp-server
        r = self.auth.client.get("legacy/applications", headers=self.auth())
        if r.status_code >= 400:
            raise ValueError("Unknown API error: " + r.text)

        aliases = set()
        conn_map = {}
        for x in r.json():
            if "config" in x:
                env = next(
                    (r["value"] for r in x["config"] if r["path"] == "env"), None
                )
                name = x["name"]
                if env is not None:
                    if env in aliases:
                        logger.warning(f"Duplicate instance alias {env}")
                        conn_map[name] = name
                        first_duplicate = conn_map.pop(env)
                        conn_map[first_duplicate] = first_duplicate
                    else:
                        conn_map[env] = name
                    aliases.add(env)
                else:
                    conn_map[name] = name
        return conn_map

    @property
    @lru_cache
    def valid_connections(self):
        """Return available resources (database API connections)"""
        logger.warning(
            "IIASA is migrating to a database infrastructure using the ixmp4 package."
            "Use `pyam.iiasa.platforms()` to list available ixmp4 databases."
        )
        return list(self._connection_map.keys())

    def connect(self, name):
        """Connect to a specific resource (database API)"""
        if name in self._connection_map:
            name = self._connection_map[name]

        valid = self._connection_map.values()
        if name not in valid:
            raise ValueError(
                f"You do not have access to instance '{name}' or it does not exist. "
                "Use `pyam.iiasa.Connection().valid_connections` for a list "
                "of accessible services."
            )

        # TODO: config will be reimplemented in conjunction with ixmp-server
        r = self.auth.client.get(
            f"legacy/applications/{name}/config", headers=self.auth()
        )
        if r.status_code >= 400:
            raise ValueError("Unknown API error: " + r.text)

        response = r.json()
        idxs = {x["path"]: i for i, x in enumerate(response)}

        self._base_url = response[idxs["baseUrl"]]["value"]
        # TODO: proper citation (as metadata) instead of link to the about page
        if "uiUrl" in idxs:
            about = "/".join([response[idxs["uiUrl"]]["value"], "#", "about"])
            logger.info(_CITE_MSG.format(name, about))

        # TODO: use API "nice-name"
        self._connected = name

    @property
    def current_connection(self):
        """Currently connected resource (database API connection)"""
        return self._connected

    @property
    def meta_columns(self):
        """Return the list of meta indicators in the connected resource"""
        url = "/".join([self._base_url, "metadata/types"])
        r = requests.get(url, headers=self.auth())
        _check_response(r)
        return pd.read_json(StringIO(r.text), orient="records")["name"]

    def _query_index(self, default_only=True, meta=False, cols=[], **kwargs):
        # TODO: at present this reads in all data for all scenarios,
        #  it could be sped up in the future to try to query a subset
        _default = "true" if default_only else "false"
        _meta = "true" if meta else "false"
        add_url = f"runs?getOnlyDefaultRuns={_default}&includeMetadata={_meta}"
        url = "/".join([self._base_url, add_url])
        r = requests.get(url, headers=self.auth())
        _check_response(r)

        # cast response to dataframe, apply filter by kwargs, and return
        runs = pd.read_json(StringIO(r.text), orient="records")
        if runs.empty:
            logger.warning("No permission to view model(s) or no scenarios exist.")
            return pd.DataFrame([], columns=META_IDX + ["version", "run_id"] + cols)

        if kwargs:
            keep = np.ones(len(runs), dtype=bool)
            for key, values in kwargs.items():
                if key not in META_IDX + ["version"]:
                    raise ValueError(f"Invalid filter: '{key}'")
                keep_col = pattern_match(pd.Series(runs[key].values), values)
                keep = np.logical_and(keep, keep_col)
            return runs[keep]
        else:
            return runs

    def index(self, default_only=True, **kwargs):
        """Return the index of models and scenarios

        Parameters
        ----------
        default_only : bool, optional
            If `True`, return *only* the default version of a model/scenario.
            If `False`, return all versions.
        **kwargs
            Arguments to filter by *model* and *scenario*, `*` can be used as wildcard.
        """
        if "default" in kwargs:
            default_only = _new_default_api(kwargs)

        cols = ["version"] if default_only else ["version", "is_default"]
        return self._query_index(default_only, **kwargs)[META_IDX + cols].set_index(
            META_IDX
        )

    def meta(self, default_only=True, run_id=False, **kwargs):
        """Return categories and indicators (meta) of scenarios

        Parameters
        ----------
        default_only : bool, optional
            If `True`, return *only* the default version of a model/scenario.
            If `False`, return all versions.
        run_id : bool, optional
            Include "run id" column
        **kwargs
            Arguments to filer by *model* and *scenario*, `*` can be used as wildcard
        """
        if "default" in kwargs:
            default_only = _new_default_api(kwargs)

        df = self._query_index(default_only, meta=True, **kwargs)

        cols = ["version"] if default_only else ["version", "is_default"]
        if run_id:
            cols.append("run_id")

        meta = df[META_IDX + cols]
        if not meta.empty and df.metadata.any():
            extra_meta = pd.DataFrame.from_records(df.metadata)
            meta = pd.concat([meta, extra_meta], axis=1)

        # remove "exclude" column when querying from an ixmp (legacy) IIASA database
        if "exclude" in meta.columns:
            meta.drop(columns="exclude", inplace=True)

        return meta.set_index(META_IDX + ([] if default_only else ["version"]))

    def properties(self, default_only=True, **kwargs):
        """Return the audit properties of scenarios

        Parameters
        ----------
        default_only : bool, optional
            If `True`, return *only* the default version of a model/scenario.
            If `False`, return all versions.
        **kwargs
            Arguments to filer by *model* and *scenario*, `*` can be used as wildcard
        """
        if "default" in kwargs:
            default_only = _new_default_api(kwargs)

        audit_cols = ["cre_user", "cre_date", "upd_user", "upd_date"]
        other_cols = ["version"] if default_only else ["version", "is_default"]
        cols = audit_cols + other_cols

        _df = self._query_index(default_only, meta=True, cols=cols, **kwargs)
        audit_mapping = {i: i.replace("_", "ate_") for i in audit_cols}

        return _df.set_index(META_IDX).rename(columns=audit_mapping)

    def models(self):
        """List all models in the connected resource"""
        return pd.Series(self._query_index()["model"].unique(), name="model")

    def scenarios(self):
        """List all scenarios in the connected resource"""
        return pd.Series(self._query_index()["scenario"].unique(), name="scenario")

    @lru_cache
    def variables(self):
        """List all variables in the connected resource"""
        url = "/".join([self._base_url, "ts"])
        r = requests.get(url, headers=self.auth())
        _check_response(r)
        df = pd.read_json(StringIO(r.text), orient="records")
        return pd.Series(df["variable"].unique(), name="variable")

    @lru_cache
    def regions(self, include_synonyms=False):
        """List all regions in the connected resource

        Parameters
        ----------
        include_synonyms : bool
            whether to include synonyms
            (possibly leading to duplicate region names for
            regions with more than one synonym)
        """
        url = "/".join([self._base_url, "nodes?hierarchy=%2A"])
        params = {"includeSynonyms": include_synonyms}
        r = requests.get(url, headers=self.auth(), params=params)
        _check_response(r)
        return self.convert_regions_payload(r.text, include_synonyms)

    @staticmethod
    def convert_regions_payload(response, include_synonyms):
        df = pd.read_json(StringIO(response), orient="records")
        if df.empty:
            return df
        if "synonyms" not in df.columns:
            df["synonyms"] = [list()] * len(df)
        df = df.astype(
            {
                "id": str,
                "name": str,
                "hierarchy": str,
                "parent": str,
                "synonyms": object,
            }
        )
        if include_synonyms:
            df = df[["name", "synonyms"]].explode("synonyms")
            return df.rename(columns={"name": "region", "synonyms": "synonym"})
        return pd.Series(df["name"].unique(), name="region")

    def _query_post(self, meta, default_only=True, **kwargs):  # noqa: C901
        def _get_kwarg(k):
            # TODO refactor API to return all models if model-list is empty
            x = kwargs.pop(k, "*" if k == "model" else [])
            return [x] if is_str(x) else x

        m_pattern = _get_kwarg("model")
        s_pattern = _get_kwarg("scenario")
        v_pattern = _get_kwarg("variable")
        r_pattern = _get_kwarg("region")

        def _match(data, patterns):
            # this is empty, return empty list which means "everything"
            if not patterns:
                return []
            # otherwise match everything
            matches = np.array([False] * len(data))
            for p in patterns:
                matches |= pattern_match(data, p)
            return data[matches].unique()

        # drop non-default runs if only default is requested
        if default_only and hasattr(meta, "is_default"):
            meta = meta[meta.is_default]

        # determine relevant run id's
        meta = meta.reset_index()
        models = _match(meta["model"], m_pattern)
        scenarios = _match(meta["scenario"], s_pattern)
        if len(models) == 0 and len(scenarios) == 0:
            runs = []
        else:
            where = np.array([True] * len(meta))
            if len(models) > 0:
                where &= meta.model.isin(models)
            if len(scenarios) > 0:
                where &= meta.scenario.isin(scenarios)
            runs = meta.run_id[where].unique().tolist()

        # get unique other values
        variables = _match(self.variables(), v_pattern)
        # pass empty list to API if all variables selected
        if len(variables) == len(self.variables()):
            variables = []
        regions = _match(self.regions(), r_pattern)
        # pass empty list to API if all regions selected
        if len(regions) == len(self.regions()):
            regions = []
        data = {
            "filters": {
                "regions": list(regions),
                "variables": list(variables),
                "runs": list(runs),
                "years": [],
                "units": [],
                "timeslices": [],
            }
        }
        return data

    def query(self, default_only=True, meta=True, **kwargs):
        """Query the connected resource for timeseries data (with filters)

        Parameters
        ----------
        default_only : bool, optional
            If `True`, return *only* the default version of a model/scenario.
            If `False`, return all versions.
        meta : bool or list, optional
            If :obj:`True`, merge all meta columns indicators
            (or subset if list is given).
        **kwargs
            Available keyword arguments include

            - model
            - scenario
            - region
            - variable

        Returns
        -------
        IamDataFrame

        Examples
        --------
        You can read from a :class:`pyam.iiasa.Connection` instance using
        keyword arguments similar to filtering an :class:`IamDataFrame`:

        .. code-block:: python

            Connection.query(
                model="MESSAGE*",
                scenario="SSP2*",
                variable=["Emissions|CO2", "Primary Energy"],
            )

        """
        if "default" in kwargs:
            default_only = _new_default_api(kwargs)

        headers = self.auth().copy()
        headers["Content-Type"] = "application/json"

        # retrieve meta (with run ids) or only index
        if meta:
            _meta = self.meta(default_only=default_only, run_id=True)
            # downselect to subset of meta columns if given as list
            if is_list_like(meta):
                # always merge 'version' (even if not requested explicitly)
                # 'run_id' is required to determine `_args`, dropped later
                _meta = _meta[list(set(meta).union(["version", "run_id"]))]
        else:
            _meta = self._query_index(default_only=default_only).set_index(META_IDX)

        # return nothing if no data exists at all
        if _meta.empty:
            return

        # retrieve data
        _args = json.dumps(self._query_post(_meta, default_only=default_only, **kwargs))
        url = "/".join([self._base_url, "runs/bulk/ts"])
        r = requests.post(url, headers=headers, data=_args)
        _check_response(r)
        # refactor returned json object to be castable to an IamDataFrame
        dtype = dict(
            model=str,
            scenario=str,
            variable=str,
            unit=str,
            region=str,
            year=int,
            value=float,
            version=int,
        )
        data = pd.read_json(StringIO(r.text), orient="records", dtype=dtype)
        cols = IAMC_IDX + ["year", "value", "subannual", "version"]
        # keep only known columns or init empty df
        data = pd.DataFrame(data=data, columns=cols)

        # check if timeseries data has subannual disaggregation, drop if not
        if "subannual" in data:
            timeslices = data.subannual.dropna().unique()
            if all([i in [-1, "Year"] for i in timeslices]):
                data.drop(columns="subannual", inplace=True)

        # define the index for the IamDataFrame
        if default_only:
            index = META_IDX
            data.drop(columns="version", inplace=True)
        else:
            index = META_IDX + ["version"]
            logger.info(f"Initializing `IamDataFrame` with non-default index {index}")

        # merge meta indicators (if requested) and cast to IamDataFrame
        if meta:
            # 'run_id' is necessary to retrieve data, not returned by default
            if not (is_list_like(meta) and "run_id" in meta):
                _meta.drop(columns="run_id", inplace=True)
            return IamDataFrame(data, meta=_meta, index=index)
        else:
            return IamDataFrame(data, index=index)


def _new_default_api(kwargs):
    # TODO: argument `default` is deprecated, remove this warning for release >= 3.1
    raise DeprecationWarning(
        "The argument `default` is deprecated, use `default_only` instead."
    )


def read_iiasa(name, default_only=True, meta=True, creds=None, **kwargs):
    """Read data from an |ixmp4| platform or an IIASA Scenario Explorer database.

    Parameters
    ----------
    name : str
        | Name of an |ixmp4| platform or an IIASA Scenario Explorer database instance.
        | Use :attr:`platforms <pyam.iiasa.platforms>` for a list of |ixmp4| platforms
          hosted by IIASA.
        | Use :attr:`valid_connections <pyam.iiasa.Connection.valid_connections>`
          for a list of available Scenario Explorer database instances.
    default_only : bool, optional
        If `True`, return *only* the default version of a model/scenario.
        If `False`, return all versions.
    meta : bool or list of strings, optional
        If `True`, include all meta categories & quantitative indicators
        (or subset if list is given).
    creds : str or :class:`pathlib.Path`, optional
        Path to a file with authentication credentials. This feature is deprecated,
        please run ``ixmp4 login <username>`` in a console instead.
    **kwargs
        Arguments for :meth:`pyam.read_ixmp4` or :meth:`pyam.iiasa.Connection.query`.

    Notes
    -----
    Credentials (username & password) are not required to access any public |ixmp4|
    or Scenario Explorer database (i.e., with Guest login).
    """
    if name in [i.name for i in ixmp4.conf.settings.manager.list_platforms()]:
        if meta is not True:
            raise NotImplementedError(
                "Reading from ixmp4 platforms requires `meta=True`"
            )
        return read_ixmp4(name, default_only=default_only, **kwargs)

    return Connection(name, creds).query(default_only=default_only, meta=meta, **kwargs)


def lazy_read_iiasa(file, name, default_only=True, meta=True, creds=None, **kwargs):
    """
    Try to load data from a local cache, failing that, loads it from an IIASA database.

    Check if the file in a given location is an up-to-date version of an IIASA
    database. If so, load it. If not, load  data from the IIASA scenario explorer
    database API and save to that location. Does not check that the previously read
    version is a complete instance of the database, so if the initial load applies a
    filter, you will read only data that passes the same filter as well as any
    additional filter you apply.

    Parameters
    ----------
    file : str or :class:`pathlib.Path`
        The location to test for valid data and save the data if not up-to-date. Must be
        either xlsx or csv.
    name : str
        | Name of an IIASA Scenario Explorer database instance.
        | Use :attr:`valid_connections <pyam.iiasa.Connection.valid_connections>`
          for a list of available instances.
    default_only : bool, optional
        If `True`, return *only* the default version of a model/scenario.
        If `False`, return all versions.
    meta : bool or list of strings, optional
        If `True`, include all meta categories & quantitative indicators
        (or subset if list is given).
    creds : str or :class:`pathlib.Path`, optional
        Path to a file with authentication credentials. This feature is deprecated,
        please run ``ixmp4 login <username>`` in a console instead.
    **kwargs
        Arguments for :meth:`pyam.read_ixmp4` or :meth:`pyam.iiasa.Connection.query`.

    Notes
    -----
    This feature does currently not support reading data from |ixmp4| platforms.

    Credentials (username & password) are not required to access any public |ixmp4|
    or Scenario Explorer database (i.e., with Guest login).
    """
    if name in [
        platform.name for platform in ixmp4.conf.settings.manager.list_platforms()
    ]:
        raise NotImplementedError(
            "The function `lazy_read_iiasa()` does not support ixmp4 platforms."
        )

    file = Path(file)
    assert file.suffix in [
        ".csv",
        ".xlsx",
    ], "We will only read and write to csv and xlsx format."
    if os.path.exists(file):
        date_set = pd.to_datetime(os.path.getmtime(file), unit="s")
        version_info = Connection(name, creds).properties()
        latest_new = np.nanmax(pd.to_datetime(version_info["create_date"]))
        latest_update = np.nanmax(pd.to_datetime(version_info["update_date"]))
        latest = pd.Series([latest_new, latest_update]).max()
        if latest < date_set:
            old_read = IamDataFrame(file)
            if kwargs:
                old_read = old_read.filter(**kwargs)
            logger.info("Database read from file")
            return old_read
        else:
            logger.info("Database out of date and will be re-downloaded")
    # If we get here, we need to redownload the database
    new_read = read_iiasa(
        name,
        meta=meta,
        default_only=default_only,
        creds=creds,
        **kwargs,
    )
    Path(file).parent.mkdir(parents=True, exist_ok=True)
    if file.suffix == ".csv":
        new_read.to_csv(file)
    else:
        new_read.to_excel(file)
    return new_read



import pandas.testing as pdt

from pyam import IamDataFrame

from . import compare


def assert_iamframe_equal(
    left: IamDataFrame,
    right: IamDataFrame,
    check_meta: bool = True,
    **kwargs,
) -> None:
    """Check that left and right IamDataFrame instances are equal.

    Parameters
    ----------
    left, right : :class:`IamDataFrame`
        Two IamDataFrame instances to be compared.
    check_meta: bool
        Whether to check that the `meta` indicators are identical.
    **kwargs
        Passed to :meth:`IamDataFrame.compare`, comparing the `data` objects.

    Raises
    ------
    AssertionError if *left* and *right* are different.

    Notes
    -----
    Columns of the *meta* attribute where all values are *nan* are ignored.
    """
    diff = compare(left, right, **kwargs)
    if not diff.empty:
        msg = "IamDataFrame.data are different: \n {}"
        raise AssertionError(msg.format(diff.head()))

    if check_meta:
        pdt.assert_frame_equal(
            left.meta.dropna(axis="columns", how="all"),
            right.meta.dropna(axis="columns", how="all"),
            check_column_type=False,
            check_dtype=False,
            check_like=True,
        )

    pdt.assert_series_equal(
        left.exclude,
        right.exclude,
    )
